﻿using Syncfusion.UI.Xaml.Grid;
using Syncfusion.UI.Xaml.Grid.Cells;
using Syncfusion.UI.Xaml.Grid.Helpers;
using Syncfusion.UI.Xaml.ScrollAxis;
using Syncfusion.UI.Xaml.Utility;
using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using System.Windows;
using System.Windows.Controls;
using System.Windows.Input;
using System.Windows.Media;
using System.Windows.Threading;

namespace EnterKeyBehavior
{
    public class AutoRowHeightHelper : IDisposable
    {

        #region Private Members

        private static double scrollbarValue = 0.0;
        protected SfDataGrid DataGrid;
        private double previoustextsize;
        int defaultLineSize = -1;

        #endregion

        #region Properties

        /// <summary>
        /// To get the VisualContainer from datagrid.
        /// </summary>
        public VisualContainer VisualContainer
        {
            get
            {
                return DataGrid.GetVisualContainer();
            }
        }
        /// <summary>
        /// To get the Default LineSize for GridCell.
        /// </summary>
        public int DefaultLineSize
        {
            get
            {
                if (defaultLineSize == -1)
                {
                    defaultLineSize = GetDefaultLineSize();
                }
                return defaultLineSize;
            }
        }

        #endregion

        #region ctor
        public AutoRowHeightHelper(SfDataGrid dataGrid)
        {
            DataGrid = dataGrid;
            dataGrid.CurrentCellEndEdit += DataGrid_CurrentCellEndEdit;
            dataGrid.CurrentCellValueChanged += DataGrid_CurrentCellValueChanged;
            dataGrid.QueryRowHeight += DataGrid_QueryRowHeight;
            dataGrid.CurrentCellActivating += DataGrid_CurrentCellActivating;
            dataGrid.SelectionController = new GridSelectionControllerExt(dataGrid);
        }

        #endregion

        #region Events and its Methods

        /// <summary>
        /// Used to Dispose (UnWire Events).
        /// </summary>
        public void Dispose()
        {
            DataGrid.CurrentCellEndEdit -= DataGrid_CurrentCellEndEdit;
            DataGrid.CurrentCellValueChanged -= DataGrid_CurrentCellValueChanged;
            DataGrid.QueryRowHeight -= DataGrid_QueryRowHeight;
            DataGrid.CurrentCellActivating -= DataGrid_CurrentCellActivating;
        }

        /// <summary>
        /// Used to Process the Navigation.
        /// </summary>
        /// <param name="sender">
        /// The SfDataGrid.
        /// </param>
        /// <param name="e">
        /// <see cref="Syncfusion.UI.Xaml.Grid.CurrentCellActivatingEventArgs"/>.
        /// The CurrentCellActivatingEventArgs.
        /// </param>       

        void DataGrid_CurrentCellActivating(object sender, CurrentCellActivatingEventArgs e)
        {
            ProcessOnCurrentCellActivating(e);
        }
        /// <summary>
        /// Update the row height when the rows or added or removed.
        /// </summary>
        /// <param name="sender">
        /// The SfDataGrid.
        /// </param>
        /// <param name="e">
        /// <see cref="Syncfusion.UI.Xaml.Grid.QueryRowHeightEventArgs"/>.
        /// The QueryRowHeightEventArgs.
        /// </param>
        private void DataGrid_QueryRowHeight(object sender, QueryRowHeightEventArgs e)
        {
            var dataGrid = sender as SfDataGrid;
            double height = 0.0;
            bool isAutoRowHeight = dataGrid.GridColumnSizer.GetAutoRowHeight(e.RowIndex, new GridRowSizingOptions(), out height);
            if (isAutoRowHeight && dataGrid.IsAddNewIndex(e.RowIndex))
            {
                if (height < dataGrid.RowHeight)
                {
                    dataGrid.GetVisualContainer().RowHeights[e.RowIndex] = dataGrid.RowHeight;
                }
            }
            else if (isAutoRowHeight && height > e.Height && height > dataGrid.RowHeight)
            {
                e.Height = height;
                e.Handled = true;
            }
        }

        /// <summary>
        /// Update the current row height and Scrolling when CurrentCellValueChanged.
        /// </summary>
        /// <param name="sender">
        /// The SfDataGrid.
        /// </param>
        /// <param name="e">
        /// <see cref="Syncfusion.UI.Xaml.Grid.CurrentCellValueChangedEventArgs"/>.
        /// The CurrentCellValueChangedEventArgs.
        /// </param>       
        private void DataGrid_CurrentCellValueChanged(object sender, CurrentCellValueChangedEventArgs e)
        {
            if (ProcessOnMultiLineTextBox(e.Column))
            {
                ProcessOnCurrentCellValueChanged(e);
            }

        }
        /// <summary>
        /// Used to check the current column is GridTextColumn.
        /// </summary>
        /// <param name="column">
        /// <see cref="Syncfusion.UI.Xaml.Grid.GridColumn"/>.
        /// The GridColumn.
        /// </param>
        /// <returns>
        /// Returns <b>true if the column is text column</b>,<b> false</b> if the column is not a text column.
        /// </returns>
        private bool ProcessOnMultiLineTextBox(GridColumn column)
        {
            var gridTextColumn = column as GridTextColumn;
            var gridMultiColumnDropDown = column as GridMultiColumnDropDownList;
            if ((gridTextColumn != null || gridMultiColumnDropDown != null) && this.DataGrid.SelectionController != null && this.DataGrid.SelectionController.CurrentCellManager != null
                && this.DataGrid.SelectionController.CurrentCellManager.CurrentCell != null &&
                this.DataGrid.SelectionController.CurrentCellManager.CurrentCell.Renderer != null)
            {
                var renderer = DataGrid.SelectionController.CurrentCellManager.CurrentCell.Renderer as GridCellRendererBase;
                TextBox textBox = renderer.CurrentCellRendererElement as TextBox;
                
                if (textBox != null && textBox.AcceptsReturn)
                {
                    return true;
                }
                //else if (textBox == null && gridMultiColumnDropDown != null)
                //{
                //    textBox = (TextBox)GridUtil.FindDescendantChildByType(renderer.CurrentCellRendererElement, typeof(TextBox));
                //    if(textBox != null && textBox.AcceptsReturn)
                //        return true;
                //}
            }
            return true;
        }

        /// <summary>
        /// Update the current cell value and Scrolling when CurrentCellValueEndEdit.
        /// </summary>
        /// <param name="sennder">
        /// The SfDataGrid.
        /// </param>
        /// <param name="e">
        /// <see cref="Syncfusion.UI.Xaml.Grid.CurrentCellEndEditEventArgs"/>.
        /// The CurrentCellEndEditEventArgs
        /// </param>     
        private void DataGrid_CurrentCellEndEdit(object sender, CurrentCellEndEditEventArgs e)
        {
            if (ProcessOnMultiLineTextBox(DataGrid.CurrentColumn))
            {
                UpdateVScrollbarValueOnCurrentCellEndEdit(e);
            }
        }

        /// <summary>
        /// Used to Process the Navigation.
        /// </summary>
        /// <param name="dataGrid">
        /// The SfDataGrid.
        /// </param>
        /// <param name="e">
        /// <see cref="Syncfusion.UI.Xaml.Grid.CurrentCellActivatingEventArgs"/>.
        /// The CurrentCellActivatingEventArgs.
        /// </param>       
        private void ProcessOnCurrentCellActivating(CurrentCellActivatingEventArgs e)
        {
            Dispatcher.CurrentDispatcher.BeginInvoke(new Action(() =>
            {
                // To get the cell current height before going to editing the cell.
                previoustextsize = DataGrid.GridColumnSizer.GetAutoCellHeight(e.CurrentRowColumnIndex);
                //Scroll the row while tab key navigation.
                DataGrid.ScrollInView(e.CurrentRowColumnIndex);
            }), DispatcherPriority.ApplicationIdle);
        }

        /// <summary>
        /// Update the current row height and Scrolling when CurrentCellValueChanged.
        /// </summary>
        /// <param name="e">
        /// <see cref="Syncfusion.UI.Xaml.Grid.CurrentCellValueChangedEventArgs"/>.
        /// The CurrentCellValueChangedEventArgs.
        /// </param>       
        private void ProcessOnCurrentCellValueChanged(CurrentCellValueChangedEventArgs e)
        {
            UpdateVScrollBarValueOnCellValueChanged(e);
        }

        /// <summary>
        /// Update the current row height and Scrolling when CurrentCellValueChanged.
        /// </summary>
        /// <param name="e">
        /// <see cref="Syncfusion.UI.Xaml.Grid.CurrentCellValueChangedEventArgs"/>.
        /// The CurrentCellValueChangedEventArgs.
        /// </param> 
        private void UpdateVScrollBarValueOnCellValueChanged(CurrentCellValueChangedEventArgs e)
        {
            RowColumnIndex rowColumnIndex = e.RowColumnIndex;
            double cellHeight = DataGrid.GridColumnSizer.GetAutoCellHeight(rowColumnIndex);
            double rowheight = 0;
            var lineInfo = VisualContainer.ScrollRows.GetVisibleLineAtLineIndex(rowColumnIndex.RowIndex);
            var scrollbarValue = VisualContainer.VScrollBar.Value;
            if (cellHeight > DataGrid.RowHeight)
            {

                var clippedOrigin = lineInfo.ClippedOrigin - lineInfo.Origin;
                var height = previoustextsize - cellHeight;
                // Get the current line size based upon caret index.
                int currentLineSize = GetCurrentLineSize(lineInfo);
                // when the grid cell is editting  in the first line of the grid cell or middle line of the grid cell or
                // last line of the grid cell , this case we need to compare the clipped size with the current line size.
                if (currentLineSize > lineInfo.ClippedSize)
                {
                    if (cellHeight == previoustextsize)
                    {
                        //when the grid cell is editing in the last visible line without "Enter" key pressed.
                        scrollbarValue += DefaultLineSize;
                    }
                    else
                    {
                        // when the grid cell is editing with "Enter" key pressed.
                        scrollbarValue += Math.Round(currentLineSize - lineInfo.ClippedSize);
                    }
                }
                else if (previoustextsize > cellHeight && lineInfo.IsClippedOrigin && lineInfo.IsClippedCorner)
                {
                    if (lineInfo.IsClippedBody)
                    {
                        // when delete the contents from middle to first in the grid cell using "BackSapce" key.
                        scrollbarValue -= Math.Round(Math.Min(clippedOrigin, height));
                    }
                    else
                    {
                        // when add the contents in the middle grid cell using "BackSapce" key.
                        scrollbarValue += Math.Round(Math.Min(clippedOrigin, height));
                    }
                }
                else if (previoustextsize > cellHeight && !lineInfo.IsClippedCorner && lineInfo.IsClippedOrigin)
                {
                    // when delete the contents from last to first in the grid cell using "BackSapce" key.
                    scrollbarValue -= Math.Round(Math.Min(clippedOrigin, height));
                }
            }
            if (cellHeight == 0 && lineInfo.IsClippedOrigin)
            {
                scrollbarValue -= (lineInfo.ClippedOrigin - lineInfo.Origin);
            }
            previoustextsize = cellHeight;
            DataGrid.GridColumnSizer.GetAutoRowHeight(rowColumnIndex.RowIndex, new GridRowSizingOptions(), out rowheight, true);
            if (rowheight < DataGrid.RowHeight)
            {
                rowheight = DataGrid.RowHeight;
            }
            VisualContainer.RowHeights[rowColumnIndex.RowIndex] = Math.Round(rowheight);
            VisualContainer.InvalidateMeasureInfo();
            VisualContainer.VScrollBar.Value = scrollbarValue;
        }

        /// <summary>
        /// Update the current cell value and Scrolling when CurrentCellValueEndEdit.
        /// </summary>
        /// <param name="e">
        /// <see cref="Syncfusion.UI.Xaml.Grid.CurrentCellEndEditEventArgs"/>.
        /// The CurrentCellEndEditEventArgs
        /// </param>     
        private void UpdateVScrollbarValueOnCurrentCellEndEdit(CurrentCellEndEditEventArgs e)
        {
            bool addNewIndex = DataGrid.IsAddNewIndex(e.RowColumnIndex.RowIndex);
            DataGrid.InvalidateRowHeight(e.RowColumnIndex.RowIndex);
            VisualContainer.InvalidateMeasureInfo();
            var lineInfo = VisualContainer.ScrollRows.GetVisibleLineAtLineIndex(e.RowColumnIndex.RowIndex);
            if (addNewIndex)
            {
                if (e.RowColumnIndex.ColumnIndex != VisualContainer.ScrollColumns.LastBodyVisibleLineIndex && lineInfo != null)
                {
                    scrollbarValue = VisualContainer.VScrollBar.Value;
                    VisualContainer.VScrollBar.Value -= lineInfo.Size - lineInfo.ClippedSize;
                }
                else if (lineInfo != null && VisualContainer.VScrollBar.Value < scrollbarValue)
                {
                    VisualContainer.VScrollBar.Value += Math.Round(DataGrid.RowHeight + lineInfo.Size);
                }
                else
                {
                    VisualContainer.VScrollBar.Value += DataGrid.RowHeight;
                }
            }
            else if (e.RowColumnIndex.ColumnIndex == VisualContainer.ScrollColumns.LastBodyVisibleLineIndex)
            {
                DataGrid.ScrollInView(new RowColumnIndex(e.RowColumnIndex.RowIndex + 1, e.RowColumnIndex.ColumnIndex));
            }
            else
            {
                DataGrid.ScrollInView(e.RowColumnIndex);
            }
        }
        /// <summary>
        /// Get the default line size of the grid cell.
        /// </summary>
        /// <returns>
        /// DefaultLineSize of the GridCell.
        /// </returns>
        private int GetDefaultLineSize()
        {
            // Increase the RowHeight for new line character
            FormattedText defaultFormattedText = new FormattedText(" ", System.Globalization.CultureInfo.CurrentCulture, FlowDirection.LeftToRight, new Typeface(DataGrid.FontFamily, new FontStyle(), FontWeights.Normal, DataGrid.FontStretch), DataGrid.FontSize, Brushes.Black);
            return (int)Math.Round(defaultFormattedText.Height);
        }

        /// <summary>
        /// Get the current line size based on the caret index.
        /// </summary>
        /// <param name="lineInfo">
        /// <see cref="Syncfusion.UI.Xaml.ScrollAxis.VisibleLineInfo"/>.
        /// The VisibleLineInfo.
        /// </param>
        /// <returns>
        /// CurrentLineSize
        /// </returns>
        private int GetCurrentLineSize(VisibleLineInfo lineInfo)
        {
            int currentLineSize = 0;
            if (DataGrid.SelectionController != null && DataGrid.SelectionController.CurrentCellManager != null
                    && DataGrid.SelectionController.CurrentCellManager.CurrentCell != null && DataGrid.SelectionController.CurrentCellManager.CurrentCell.Renderer != null)
            {
                var clippedOrigin = lineInfo.ClippedOrigin - lineInfo.Origin;
                var renderer = DataGrid.SelectionController.CurrentCellManager.CurrentCell.Renderer as GridCellRendererBase;
                TextBox textBox;
                if (renderer.CurrentCellRendererElement is CustomSfMultiColumnDropDown)
                {
                    textBox = (renderer.CurrentCellRendererElement as CustomSfMultiColumnDropDown).TextEditor;
                }
                else
                    textBox = (TextBox)renderer.CurrentCellRendererElement;
                int lineIndex = textBox.GetLineIndexFromCharacterIndex(textBox.CaretIndex);
                currentLineSize = (lineIndex + 1) * 16;
                if (lineInfo.IsClippedOrigin)
                {
                    currentLineSize -= (int)(clippedOrigin);
                }
            }
            return currentLineSize;
        }

        #endregion


        #region InnerClass - SelectionControllerExt

        /// <summary>
        /// Implement the multiline using Selectioncontroller
        /// </summary>
        public class GridSelectionControllerExt : GridSelectionController
        {
            #region Private Members

            public bool IsLastLineIndex = false;
            int defaultLineSize = -1;
            protected SfDataGrid Datagrid;

            #endregion

            #region ctor
            public GridSelectionControllerExt(SfDataGrid dataGrid) : base(dataGrid)
            {
                Datagrid = dataGrid;
            }
            #endregion

            #region Properties

            /// <summary>
            /// To get the Default LineSize for GridCell.
            /// </summary>
            public int DefaultLineSize
            {
                get
                {
                    if (defaultLineSize == -1)
                    {
                        defaultLineSize = GetDefaultLineSize();
                    }
                    return defaultLineSize;
                }
            }
            #endregion

            #region Methods

            /// <summary>
            /// Get the default line size of the grid cell.
            /// </summary>
            /// <returns>
            /// DefaultLineSize of the GridCell.
            /// </returns>
            private int GetDefaultLineSize()
            {
                // Increase the RowHeight for new line character
                FormattedText defaultFormattedText = new FormattedText(" ", System.Globalization.CultureInfo.CurrentCulture, FlowDirection.LeftToRight, new Typeface(DataGrid.FontFamily, new FontStyle(), FontWeights.Normal, DataGrid.FontStretch), DataGrid.FontSize, Brushes.Black);
                return (int)Math.Round(defaultFormattedText.Height);
            }

            /// <summary>
            /// Customize the key navigation
            /// </summary>
            /// <param name="args"></param>
            protected override void ProcessKeyDown(KeyEventArgs args)
            {
                if (DataGrid.SelectionController.CurrentCellManager.CurrentCell.IsEditing)
                {
                    if (args.Key == Key.Down || args.Key == Key.Up)
                        UpdateScrollBar(DataGrid, CurrentCellManager.CurrentRowColumnIndex, args);
                    if ((args.Key == Key.Enter) || (args.Key == Key.Left)
                        || (args.Key == Key.Right) || (args.Key == Key.Up) || (args.Key == Key.Down))
                    {
                        args.Handled = false;
                        return;
                    }
                }
                base.ProcessKeyDown(args);
            }

            /// <summary>
            /// Get the current line size based on the caret index.
            /// </summary>
            /// <param name="lineInfo">
            /// <see cref="Syncfusion.UI.Xaml.ScrollAxis.VisibleLineInfo"/>.
            /// The VisibleLineInfo.
            /// </param>
            /// <returns>
            /// CurrentLineSize
            /// </returns>
            private int GetCurrentLineSize(VisibleLineInfo lineInfo)
            {
                int currentLineSize = 0;
                if (this.DataGrid.SelectionController != null && this.DataGrid.SelectionController.CurrentCellManager != null
                        && this.DataGrid.SelectionController.CurrentCellManager.CurrentCell != null && this.DataGrid.SelectionController.CurrentCellManager.CurrentCell.Renderer != null)
                {
                    var clippedOrigin = lineInfo.ClippedOrigin - lineInfo.Origin;
                    var renderer = this.DataGrid.SelectionController.CurrentCellManager.CurrentCell.Renderer as GridCellRendererBase;
                    var textBox = (TextBox)renderer.CurrentCellRendererElement;
                    int lineIndex = textBox.GetLineIndexFromCharacterIndex(textBox.CaretIndex);
                    int lastLineIndex = textBox.GetLastVisibleLineIndex();
                    currentLineSize = (lineIndex + 1) * DefaultLineSize;
                    IsLastLineIndex = lineIndex == lastLineIndex ? IsLastLineIndex = true : IsLastLineIndex = false;
                    if (lineInfo.IsClippedOrigin)
                    {
                        currentLineSize -= (int)(clippedOrigin);
                    }
                }
                return currentLineSize;
            }

            /// <summary>
            /// Update VScrollBar value when navigation is  applied inside the cell.
            /// </summary>
            /// <param name="dataGrid">
            /// The SfDataGrid.
            /// </param>
            /// <param name="rowColumnIndex">
            /// The RowColumnIndex.
            /// </param>
            /// <param name="args">
            /// The KeyBoard Event args
            /// </param>
            private void UpdateScrollBar(SfDataGrid dataGrid, RowColumnIndex rowColumnIndex, KeyEventArgs args)
            {
                VisualContainer visualContainer = dataGrid.GetVisualContainer();
                var lineInfo = visualContainer.ScrollRows.GetVisibleLineAtLineIndex(rowColumnIndex.RowIndex);
                var scrollbarValue = visualContainer.VScrollBar.Value;
                var clippedOrigin = lineInfo.ClippedOrigin - lineInfo.Origin;
                // Get the current line size based upon caret index.
                int currentLineSize = GetCurrentLineSize(lineInfo);
                bool isNeedUpdateScrollValue = args.Key == Key.Down ? !IsLastLineIndex : true;
                if (isNeedUpdateScrollValue)
                {
                    if (args.Key == Key.Down)
                    {
                        currentLineSize += DefaultLineSize;
                    }
                    else if (args.Key == Key.Up && currentLineSize > 16)
                    {
                        currentLineSize -= DefaultLineSize;
                    }
                    if (currentLineSize >= lineInfo.ClippedSize)
                    {
                        scrollbarValue += Math.Round(currentLineSize - lineInfo.ClippedSize);
                    }
                    else if (currentLineSize < DefaultLineSize && currentLineSize > 0)
                    {
                        scrollbarValue -= DefaultLineSize;
                    }
                }
                visualContainer.VScrollBar.Value = scrollbarValue;
                visualContainer.InvalidateMeasureInfo();
            }
        }
        #endregion

        #endregion
    }
}
