﻿using System;
using System.Collections.Generic;
using System.Collections.ObjectModel;
using System.ComponentModel;
using System.Linq;
using System.Runtime.CompilerServices;
using System.Text;
using System.Threading.Tasks;

namespace EnterKeyBehavior
{
    public abstract class BindableBase : INotifyPropertyChanged
    {
        public event PropertyChangedEventHandler PropertyChanged;


        protected bool SetProperty<T>(ref T storage, T value, [CallerMemberName] String propertyName = null)
        {
            if (object.Equals(storage, value)) return false;

            storage = value;
            this.OnPropertyChanged(propertyName);
            return true;
        }

        protected void OnPropertyChanged([CallerMemberName] string propertyName = null)
        {
            var eventHandler = this.PropertyChanged;
            if (eventHandler != null)
            {
                eventHandler(this, new PropertyChangedEventArgs(propertyName));
            }
        }
    }
    
    public class ProductModel
    {
        public string Name { get; set; }
        public string Code { get; set; }
    }

    public class OrderModel : BindableBase
    {
        private ProductModel product;
        public ProductModel Product
        {
            get { return product; }
            set { this.SetProperty(ref product, value); }
        }

        private string description;
        public string Description
        {
            get { return description; }
            set { this.SetProperty(ref description, value); }
        }

        private double price;
        public double Price
        {
            get { return price; }
            set { this.SetProperty(ref price, value); }
        }
    }

    public class ProductsViewModel : BindableBase
    {
        public ProductsViewModel()
        {
            Load();
        }

        private ObservableCollection<OrderModel> orders = new ObservableCollection<OrderModel>();
        public ObservableCollection<OrderModel> Orders
        {
            get { return orders; }
            set { this.SetProperty(ref orders, value); }
        }

        private ObservableCollection<ProductModel> products = new ObservableCollection<ProductModel>();
        public ObservableCollection<ProductModel> Products
        {
            get { return products; }
            set { this.SetProperty(ref products, value); }
        }

        public void Load()
        {
            for (int index = 0; index < 1000; index++)
            {
                Products.Add(new ProductModel() { Name = "Product name with a very long string " + index, Code = "C" + index });
            }


            for (int index = 0; index < 5; index++)
            {
                Orders.Add(new OrderModel() { Description = "Description with a very long string " + index, Price = 1000 + index, Product = Products[index] });
            }
        }
    }
     
}
